"""

  Usability application class  
  Copyright 2005 Mikko Ohtamaa

  See class comments for more details

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as described
  in licence.txt

"""

# Zope imports
from AccessControl import getSecurityManager, ClassSecurityInfo, Unauthorized
from DateTime import DateTime

# Plone imports
from Products.Archetypes.public import *
from Products.ATContentTypes.content.base import ATCTOrderedFolder
from Products.ATContentTypes.content.base import updateActions
from Products.ATContentTypes.content.folder import ATFolderSchema, ATFolder
from Products.CMFCore.utils import getToolByName
from Products.CMFCore.CMFCorePermissions import ModifyPortalContent, DeleteObjects, ListFolderContents, View
from Products.URLWidget.URLWidget import URLWidget
from Products.Usability.helpers import addPermissionsForRole, doNotAcquirePermissions, forceWorkflowAction
from Products.Usability.Issue import Issue
from Products.Usability.UsabilityTool import UsabilityTool



from config import PROJECTNAME

from Permissions import *

schema = ATFolderSchema.copy() +  Schema(
    # field list
    (    
        TextField('description',
                  searchable=True,
                  required=True,                  
                  default_output_type='text/plain',
                  allowable_content_types=('text/plain',), 
                  widget=TextAreaWidget(label='Description',
                      description="What application does",
                      allow_file_upload = False,
                  ),
        ),
                    
        StringField('homepage',
                    searchable=True,
                    widget=URLWidget(
                        description="URL to application www site",
                    ),                    
            ),

        TextField('contact',
                   searchable = True,
                    widget=TextAreaWidget(
                        description="How to contact to developers to report usability issues",
                  ),
            ),
                  
        ImageField('screenshot',
            widget=ImageWidget(
                variants={ 
                    "thumb"  : (128,128),
                    "normal" : (1280,1024),
                },
                description="Screenshot from the application",
            ),
            
            )
    )
    #other schema items (none)
)

# Rename title field to be more meaningful
schema['title'].widget.label = "Application name"

class Application(ATFolder):
    """ Application in usability database
    
    Some tweaking have been done to
    - Ensure that issue id's are format we want
    - Issues don't appear before save is hit in     
    """

    schema = schema
        
    # Generate user friendly id from title during creation
    _at_rename_from_title = True
    
    # Allow insertion only to ApplicationCategory
    global_allow = False
    
    meta_type      = 'Application'
    archetype_name = 'Application'
    content_icon   = 'application_icon.gif'
    
    # - Use custom view (inc. both folder listing and fields)    
    default_view   = 'application_view'
    
    filter_content_types = True
    allowed_content_types = ['Issue',]
    
    typeDescription= 'One application in usability issue database'    
    
    security = ClassSecurityInfo()
    
    # - Use custom permissions for view & edit
    # - Hide contents tab from clueless users
    actions = updateActions(ATFolder,
        (
        {
        'id'          : 'view',
        'name'        : 'View',
        'action'      : 'string:${object_url}',
         },         
                     
         {
        'id'          : 'edit',
        'name'        : 'Edit',
        'action'      : 'string:${object_url}/edit',
         },         
         
         # Hide Properties tab from ordinal users
         { 'id': 'metadata',
           'name': 'Properties',
           'action': 'string:${object_url}/base_metadata',
           'permissions': (MANAGE_USABILITY_ITEMS_PERMISSION,),
         },         
         
         # Handled specially in disiplayContentsTab()
         #{
         #'id'          : 'folderContents',
         #'name'        : 'Contents',
         #'action'      : 'string:${folder_url}/folder_contents',
         #'permissions' : (MANAGE_USABILITY_ITEMS_PERMISSION,),         
         #'condition'   : None,
         #},
        )
    )

    def initializeArchetype(self, **kwargs):
        """ Set voting enabled initially
        
        enableVoting must be enabled in the parent folder for its children items

        called by the generated addXXX factory in types tool """
        ATFolder.initializeArchetype(self, **kwargs)
        setattr(self, "enableRatings", True)
        setattr(self, "enableVoting", True)
                
        #doNotAcquirePermissions(self, (ModifyPortalContent, DeleteObjects, ListFolderContents,
        #                                EDIT_USABILITY_ITEMS_PERMISSION, 
        #                                ADD_ISSUES_PERMISSION, ))
                        
        # set default rights for other members as the creator
        #addPermissionsForRole(self, "Anonymous", (VIEW_USABILITY_ITEMS_PERMISSION,))
        addPermissionsForRole(self, "Member", (ADD_ISSUES_PERMISSION,))
        addPermissionsForRole(self, "Manager", (MANAGE_USABILITY_ITEMS_PERMISSION,))
                                        
        # Enable RSS change feeds for new issues by default
        # syndicationTool is CMFDefault/SyndicationTool
        syndicationTool = getToolByName(self, 'portal_syndication')
        try:
            # Syndication might be disabled, so we fail will silently
            # TODO: Not too bad things might happen
            syndicationTool.enableSyndication(self)
        except:
            pass
        
            
    security.declareProtected(ADD_ISSUES_PERMISSION, 'createObject')
    def createObject(self, RESPONSE=None):
        """ Create a new issue whose local and site wide id is generated in non-standard way
        
        Give U###### format id for the created item.
        This id is used as Zope folder level object id (id, oid) and 
        Archetypes site wide unique id (uid).        
        
        Bad things happen, if generated ids are not site wide unique!
        """                
                
        usabilityTool = getToolByName(self, 'portal_usability')
        
        # create string UI + unique number (a running counter)
        id = "UI" + str(usabilityTool.nextIssueId())
        
        # Create a new object with an object id
        issue = Issue(id)
        
        # Force Archetypes UID for the object.
        # Otherwise it is generated through ReferenceEngine / make_uuid call.
        issue._at_uid = id
        
        # assing created object to this (self) container
        self._setObject(id, issue)
        issue = issue.__of__(self)
        
        # Initialize Archetypes stuff for the created object
        issue.initializeArchetype()
                
        # Go to the action which is responsible for editing a newly create item
        if RESPONSE != None:
            RESPONSE.redirect(issue.absolute_url() + '/issue_basic_edit')
            
    def displayContentsTab(self):
        """ Display Contents tab only for managers 
        
        Reduce information flow for clueless users
        
        Seems that for contents tab all permissions checks are ignored.
        I couldn't get contents hidden by settings Listing denided
        or by modifying folderListing action. It is specially handled.
        """
        return getSecurityManager().checkPermission(MANAGE_USABILITY_ITEMS_PERMISSION, self)

registerType(Application, PROJECTNAME)
