"""

    Installs Usability product
    
    Copyright 2005 Mikko Ohtamaa
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as described
    in licence.txt

    Tasks:

    - Installs skin files for custom code
    - Adds new content types: categories, application, issue
    - Adds new roles: Usability Manager, Usability Editor
    - Install issue & application category workflows
    
    Sometimes installer may crash so that ZODB is left to an inconsistent state:
    Rerunning installer causes AttributeError exceptions: there seems to be ghost object
    in some catalogs preventing a clean run. This can be fixed by going to portal_catalog
    tool and hitting "Update Catalog" from Advanced tab.    

"""
import transaction 
from StringIO import StringIO
from zLOG import *
from Products.Archetypes.public import listTypes
from Products.Archetypes.Extensions.utils import installTypes, install_subskin
from Products.ATRatings.RatingsTool import RatingsTool
from Products.ATVocabularyManager.types import simple
from Products.CMFCore import MembershipTool
from Products.CMFCore.permissions import AccessContentsInformation
from Products.CMFCore.permissions import ModifyPortalContent
from Products.CMFCore.permissions import RequestReview
from Products.CMFCore.permissions import ReviewPortalContent
from Products.CMFCore.permissions import View
from Products.CMFCore.permissions import DeleteObjects
from Products.CMFCore.permissions import ListFolderContents
from Products.CMFCore.WorkflowTool import addWorkflowFactory, _makeWorkflowFactoryKey
from Products.CMFCore.utils import getToolByName, manage_addTool
from Products.CMFPlone.FolderWorkflow import createFolderWorkflow
from Products.CMFCalendar.EventPermissions import ChangeEvents
from Products.DCWorkflow.DCWorkflow import DCWorkflowDefinition
from Products.DCWorkflow.Guard import Guard
from Products.DCWorkflow.Default import *
from Products.Usability.SampleContent import installContent
from Products.Usability.UsabilityTool import UsabilityTool
from Products.Usability.helpers import addPermissionsForRole

from Products.Usability.config import * 
from Products.Usability.Permissions import * 



# Which tools need installation
# List of tool ids
tools = (UsabilityTool.id,)

# Debug global reference to output buffer
out_holder = None

def log(str):
    LOG("Usability Installer", ERROR, str)

def addRole(portal, user, role):
    """ Modify user roles by directly calling acl_users """
    roles = user.getRoles() + (role,)
    portal.acl_users._doChangeUser(user.getUserName(), user.getPassword(), roles, user.getDomains())

    
def getGroupById(portal, id):
    """Returns the portal_groupdata-ish object for a group corresponding
    to this id."""
    if id==None:
        return None
    prefix = portal.acl_users.getGroupPrefix()
    return portal.acl_users.getGroup(id, prefixed=id.startswith(prefix)) 
    
def setupTools(portal, out):
    """ Add IssueVoteTool to portal tools

    Check if we have already tool by installed (should not happen)

    """
    
    # Install UsabilityTool tool             
    if not hasattr(portal, UsabilityTool.id):
        portal.manage_addProduct[PROJECTNAME].manage_addTool(UsabilityTool.meta_type)        
        print >> out, 'tool installed'
        
def createIssueFeatures(portal, out):
    """ Creates default issue features 
    
    Calls ATVocabularyManager product, creates issue_features vocabulary
    """
    vocabTool = portal.portal_vocabularies
    
    if not FEATURE_VOCABULARY in vocabTool.objectIds():
        vocabTool.invokeFactory(type_name = "SimpleVocabulary", id = FEATURE_VOCABULARY)
        
    featureVocab = vocabTool[FEATURE_VOCABULARY]
    
    if not "design_error" in featureVocab.objectIds():
        featureVocab.addTerm("design_error", "Design error")
    
    if not "breaking_guidelines" in featureVocab.objectIds():
        featureVocab.addTerm("breaking_guidelines", "Guidelines are not followed")
                      
    if not "unfamiliar_vocabulary" in featureVocab.objectIds():
        featureVocab.addTerm("unfamiliar_vocabulary", "Unfamiliar vocabulary or bad analogue")
        
    if not "extra_work" in featureVocab.objectIds():
        featureVocab.addTerm("extra_work", "Extra work for a user")        

    if not "missing_proceeding_hints" in featureVocab.objectIds():
        featureVocab.addTerm("missing_proceeding_hints", "Missing proceeding hints")

    if not "repeatly_interrupted" in featureVocab.objectIds():
        featureVocab.addTerm("repeatly_interrupted", "Repeatly interrupted")        
                
                                
            
def createIssueWorkflow(id):
    """ Creates Usability issue workflow for the portal
    
    There is no state transition control 
    though it is possible to implement
    
    """            
    flow=DCWorkflowDefinition(id)    
        
    # Install Issue workflow
    flow.title = "Usability Issue Workflow"
    
    # All possible states in usability issue 
    all_states = ( "in_progress", "author_notified", "fixed", "permanent", "sealed" )
        
    # ...and transitions    
    all_transitions = ( "to_in_progress", "to_author_notified", "to_fixed", "to_permanent", "seal", "reopen" )
    
        
    # this var name is hardcoded in global_defines.pt
    flow.variables.setStateVar('review_state')    
    flow.initial_state = "in_progress"
    flow.permissions+=(EDIT_USABILITY_ITEMS_PERMISSION, )
        
    flow.states.addState('in_progress')    
    inprogress = flow.states["in_progress"]
    inprogress.title = "In progress"
    inprogress.description = "The issue is initially reported here"
    inprogress.transitions = ( "to_author_notified", "to_fixed", "to_permanent", "seal" )
        
    flow.states.addState('author_notified')    
    state = flow.states["author_notified"]
    state.title = "Author notified"
    state.description = "Application authors are aware of the issue"
    state.transitions = ( "in_progress", "to_fixed", "to_permanent", "seal" )

    flow.states.addState('fixed')    
    state = flow.states["fixed"]
    state.title = "Fixed"
    state.description = "Issue is fixed in a new version"
    state.transitions = ( "to_author_notified", "in_progress", "to_permanent", "seal" )
    
    flow.states.addState('permanent')    
    state = flow.states["permanent"]
    state.title = "Permanent issue"
    state.description = "Authors have expressed that they are not going to fix the issue "
    state.transitions = ( "to_author_notified", "to_fixed", "in_progress", "seal" )
  
    flow.states.addState('sealed')    
    state = flow.states["sealed"]
    state.title = "Sealed"
    state.description = "The issue has been frozen for editing"
    state.transitions = ('reopen',)
        
    # Only users with management permissions can do state transitions
    managerTransitionGuard = Guard()
    managerTransitionGuard.permissions = (MANAGE_USABILITY_ITEMS_PERMISSION,)                
        
    flow.transitions.addTransition("to_in_progress")
    trans = flow.transitions["to_in_progress"]
    trans.title = "To in progress"
    trans.actbox_name = "To in progress"
    trans.description = "Set state to 'in progress'"
    trans.new_state_id = "in_progress"
    trans.guard = None
    trans.actbox_url='%(content_url)s/content_change_state_form'
 
    flow.transitions.addTransition("to_author_notified")
    trans = flow.transitions["to_author_notified"]
    trans.title = "To author notified"
    trans.actbox_name = "To author notified"
    trans.description = "Set state to 'author notified'"
    trans.new_state_id = "author_notified"
    trans.guard = None
    trans.actbox_url='%(content_url)s/content_change_state_form'    
    
    flow.transitions.addTransition("to_fixed")
    trans = flow.transitions["to_fixed"]
    trans.title = "To fixed"
    trans.actbox_name = "To fixed"
    trans.description = "Set state to 'fixed'"
    trans.new_state_id = "fixed"
    trans.guard = None
    trans.actbox_url='%(content_url)s/content_change_state_form'    
    
    flow.transitions.addTransition("to_permanent")
    trans = flow.transitions["to_permanent"]
    trans.title = "To permanent"
    trans.actbox_name = "To permanent"
    trans.description = "Set state to 'permanent'"
    trans.new_state_id = "permanent"
    trans.guard = None
    trans.actbox_url='%(content_url)s/content_change_state_form'    

    flow.transitions.addTransition("seal")
    trans = flow.transitions["seal"]
    trans.title = "Seal"
    trans.actbox_name = "Seal"
    trans.description = "Seal issue to prevent editing"
    trans.new_state_id = "sealed"
    trans.guard = managerTransitionGuard
    trans.actbox_url='%(content_url)s/content_change_state_form'            
    
    flow.transitions.addTransition("reopen")
    trans = flow.transitions["reopen"]
    trans.title = "Reopen"
    trans.actbox_name = "Reopen"
    trans.description = "Make issue editable again"
    trans.new_state_id = "in_progress"
    trans.guard = managerTransitionGuard
    trans.actbox_url='%(content_url)s/content_change_state_form'    
    
    # In inprogress state, all members can edit the item. In sealed state, 
    # it is only possible for managers.
    # EDIT_USABILITY_ITEMS_PERMISSION set in any other point of code,
    # it is set only here.    
    
    # Member gains back its permissions
    flow.states.in_progress.setPermission(EDIT_USABILITY_ITEMS_PERMISSION, False, ('Member','Manager', USABILITY_MANAGER_ROLE,USABILITY_EDITOR_ROLE,))
    flow.states.author_notified.setPermission(EDIT_USABILITY_ITEMS_PERMISSION, False, ('Member','Manager', USABILITY_MANAGER_ROLE,USABILITY_EDITOR_ROLE,))
    flow.states.fixed.setPermission(EDIT_USABILITY_ITEMS_PERMISSION, False, ('Member','Manager', USABILITY_MANAGER_ROLE,USABILITY_EDITOR_ROLE,))
    flow.states.permanent.setPermission(EDIT_USABILITY_ITEMS_PERMISSION, False, ('Member','Manager', USABILITY_MANAGER_ROLE,USABILITY_EDITOR_ROLE,))
    flow.states.sealed.setPermission(EDIT_USABILITY_ITEMS_PERMISSION, False, ('Manager', USABILITY_MANAGER_ROLE,USABILITY_EDITOR_ROLE,))

    return flow
    
def createApplicationWorkflow(id):
    """ Creates folder_workflow compatible workflow
    
    - Initial state is published instead of visible
    - Registered users can modify contents
    """
    flow = createFolderWorkflow(id)
    flow.setProperties(title='Application Workflow')
    flow.states.setInitialState('published')
    
    p_access = AccessContentsInformation
    p_modify = ModifyPortalContent
    p_request = RequestReview
    p_review = ReviewPortalContent
    p_view = View

    r_anon = 'Anonymous'
    r_manager = 'Manager'
    r_reviewer = 'Reviewer'
    r_owner = 'Owner'
    r_member = 'Member'
    
    sdef = flow.states['published']    
    sdef.setPermission(p_access, 1, (r_anon, r_manager))
    sdef.setPermission(p_view, 1, (r_anon, r_manager))
    sdef.setPermission(p_modify, 0, (r_manager, r_member))
        
    return flow
    

def createCategoryWorkflow(id):
    """ Creates folder_workflow compatible workflow
    
    - Initial state is published instead of visible    
    - Only private and published states supported:
      in other states normal users can do bad things, e.g. delete items
    """
    flow = createFolderWorkflow(id)
    flow.setProperties(title='Application Category Workflow')
    flow.states.setInitialState('published')
    
    
    flow.permissions+=(
        ADD_ISSUES_PERMISSION, 
        EDIT_USABILITY_ITEMS_PERMISSION, 
        VIEW_USABILITY_ITEMS_PERMISSION,
        MANAGE_USABILITY_ITEMS_PERMISSION,
        DeleteObjects,
        ListFolderContents)
    
    p_access = AccessContentsInformation
    p_modify = ModifyPortalContent
    p_request = RequestReview
    p_review = ReviewPortalContent
    p_view = View
    p_delete = DeleteObjects

    r_anon = 'Anonymous'
    r_manager = 'Manager'
    r_reviewer = 'Reviewer'
    r_owner = 'Owner'
    r_member = 'Member'
    
    # This is the state where all usability categories should
    # be all the time
    sdef = flow.states['published']    
    sdef.setPermission(p_access, True, (r_anon, r_manager))
    sdef.setPermission(p_view, True, (r_anon, r_manager))
    sdef.setPermission(p_modify, False, (r_manager,))
    sdef.setPermission(p_delete, False, (r_manager,))
    sdef.setPermission(ADD_ISSUES_PERMISSION, False, (r_member,r_manager,))
    sdef.setPermission(VIEW_USABILITY_ITEMS_PERMISSION, False, (r_anon,r_manager,))
    sdef.setPermission(ADD_ISSUES_PERMISSION, False, (r_member,r_manager,))
    sdef.setPermission(EDIT_USABILITY_ITEMS_PERMISSION, False, (r_manager,))
    sdef.setPermission(MANAGE_USABILITY_ITEMS_PERMISSION, False, (r_manager,))    
    sdef.setPermission(ListFolderContents, False, (r_manager,))
    
        
    return flow
         
                                        
def setupWorkflow(workflowTool, flow_id, flow_title, defaultPortalTypes, creatorFunction):                                        
    """ """
    
    addWorkflowFactory(creatorFunction, id=flow_id, title=flow_title)        
    
    # Plone 2.1 fix: pt_types is a tuple instead of string
    workflowTool.setChainForPortalTypes(defaultPortalTypes, flow_id)
    
    workflow_type=_makeWorkflowFactoryKey(creatorFunction, id=flow_id, title=flow_title)          
    workflowTool.manage_addWorkflow(id=flow_id, workflow_type=workflow_type)
            
    
def setupWorkflows(portal, out):
    """ Create workflows for this product
    
    Parameters:
        - portal: Plone object
        - out: StringIO getting installer log output 
    """
    
    setupWorkflow(portal.portal_workflow, "issue_workflow", "Usability issue workflow", 
                   ("Issue",), createIssueWorkflow)


    setupWorkflow(portal.portal_workflow, "application_workflow", "Application workflow", 
                   ("Application",), createApplicationWorkflow)
                   
    setupWorkflow(portal.portal_workflow, "application_category_workflow", "Application category workflow", 
                   ("ApplicationSuperCategory","ApplicationSubCategory",), createCategoryWorkflow)
                          
    # Refresh magic 
    # (not sure if required)     
    portal.portal_catalog.refreshCatalog()
    portal.portal_workflow.updateRoleMappings()
    
    out.write("Succesfully installed new workflows")
    
        
def install(self):
    """ Quick install script
    
    This script is executed when user chooses quickinstall in Zope Management Interface
    or Plone Management Page 
    
    @param portal Plone portal where are installing to
    """
    
    portal = self
    
    # Create buffer which logs messages during installation
    out = StringIO()

    # Register Archetypes types in the portal
    installTypes(self, out, listTypes(PROJECTNAME), PROJECTNAME)

    # Register layout files in the portal
    install_subskin(self, out, GLOBALS)          
        
    # Register roles used in dealing with usability issues
    try:
        portal._addRole(USABILITY_MANAGER_ROLE)
        portal._addRole(USABILITY_EDITOR_ROLE)
        portal._addRole(USABILITY_VIEWER_ROLE)        
    except:
        # TODO: Is exception thrown if roles exist?
        pass
    
    # Add all manager user to have usability manager role
    membership = portal.portal_membership
    memberdata = portal.portal_memberdata

    managers = membership.searchForMembers(roles=('Manager', 'Owner'))        
    for user in managers:
        addRole(portal, user, USABILITY_MANAGER_ROLE)
        
        
    #addPermissionsForRole(portal, "Member", (VIEW_USABILITY_ITEMS_PERMISSION, ADD_ISSUES_PERMISSION, EDIT_USABILITY_ITEMS_PERMISSION,))
    addPermissionsForRole(portal, "Manager", (VIEW_USABILITY_ITEMS_PERMISSION, ADD_ISSUES_PERMISSION, MANAGE_USABILITY_ITEMS_PERMISSION,EDIT_USABILITY_ITEMS_PERMISSION,))
    
    # Register ZMI tools in the portal
    setupTools(self, out)
        
    
    # Add issue management workflow
    setupWorkflows(portal, out)
                                                       
    # Create issue feature vocabulary
    createIssueFeatures(portal, out)
    
    # Place some sample content into the portal
    # installContent(portal, out)
        
    out.write("Successfully installed %s." % PROJECTNAME)
    return out.getvalue()

    
    