"""

  Usability issue item class  
  Copyright 2005 Mikko Ohtamaa

  See class comments for more details

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as described
  in licence.txt

"""

# Zope imports
from AccessControl import getSecurityManager, ClassSecurityInfo, Unauthorized

# Plone imports
from Products.Archetypes.ExtensibleMetadata import ExtensibleMetadata
from Products.ATContentTypes.content.base import ATCTContent
from Products.ATContentTypes.content.base import updateActions, updateAliases
from Products.ATContentTypes.content.schemata import ATContentTypeSchema
from Products.ATRatings.IVoteable import IVoteable
from Products.CMFCore.utils import getToolByName
from Products.CMFCore import CMFCorePermissions
from Products.CMFCore.CMFCorePermissions import View, ManageProperties, ListFolderContents, ModifyPortalContent
from Products.TypedReferenceWidget.TypedReferenceWidget import TypedReferenceWidget
from Products.TypedReferenceWidget.UIDListValidator import UIDListValidator
from Products.Usability.helpers import addPermissionsForRole, doNotAcquirePermissions

from Products.Archetypes.public import \
    StringField, \
    IntegerField, \
    ReferenceField, \
    TextField, \
    TextAreaWidget, \
    IntegerWidget, \
    SelectionWidget, \
    MultiSelectionWidget, \
    ReferenceWidget, \
    StringWidget, \
    RichWidget, \
    ImageField, \
    ImageWidget, \
    KeywordWidget, \
    BaseContent, \
    BaseSchema, \
    Schema, \
    registerType
    
# local imports

from Permissions import *
from config import PROJECTNAME, ISSUE_LEVELS, ISSUE_STATUSES, ISSUE_FEATURES, FEATURE_VOCABULARY

def showIn(cond, fieldName):
    """ Generate view condition statement for Issue widgets
        
    @param cond Views where item should be shown
       
    """
    return "python: modules['Products.Usability'].isRequestView(object, '" + str(cond) + "') "

def showNotIn(cond):
    return "python: not modules['Products.Usability'].isRequestView(object, '" + str(cond) + "')"
    
def issueFieldVisibility(object, visibleViews, getterName):
    """ Helper function to decide field visibility
    
    Done in skins/TAL
    """       
    pass

#
# This schema defines issue fields and their editing widgets
#
schema = ATContentTypeSchema.copy() + Schema(
    # field list
    (
        StringField('name',
            searchable=True,
            required=True,
            widget=StringWidget(
                label="Name",
                visible = {'edit': 'visible', 'view': 'invisible'},
            ),
            # this is needed because Title() returns title in form ui + title
            # Title() value is used by folder listing directly.
            # Changing title accessor doesn't affect it.
        ),

        StringField('severity',
            searchable=1,
            required=1,
            default='cosmetic',
            enforceVocabulary=False,
            vocabulary=ISSUE_LEVELS,
            widget=SelectionWidget(
                label='Severity',
                description="Effect on user experience",
                format='select'
            ),
        ),

        ReferenceField("relatedTo",
            relationship="reference",
            multiValued=1,            
            allowed_types=("Issue",),
            addable="False",
            validators=UIDListValidator("RelatedToValidator"),
            widget=TypedReferenceWidget(       
                label="Related issues",
                description="Comma separated list of related issue ids",
                change_case = "upper",
                condition=showNotIn('issue_basic_edit'),
                destination_types="Issue",
                destination="/"
            ),
        ),

        TextField('issueDescription',
            searchable=True,
            required=True,
            default_output_type='text/plain',
            allowable_content_types=('text/plain',),
            # Store data in Description meta field
            # use Description's accessors
            accessor="Description",
            edit_accessor="Description",
            mutator="setDescription",                        
            widget=TextAreaWidget(
                label='Description',
                description="Describe the problem",
                allow_file_upload=False,
                ),
            ),

        TextField('rootCause',
            searchable=True,
            required=False,
            default_output_type='text/plain',
            allowable_content_types=('text/plain',),
            widget=TextAreaWidget(
                label='Root cause',
                allow_file_upload=False,
                condition=showNotIn('issue_basic_edit'),                
                description="What is causing this problem",
                ),
            ),

        TextField('solution',
            searchable=True,
            required=False,
            default_output_type='text/plain',
            allowable_content_types=('text/plain',),
            widget=TextAreaWidget(
                label='Proposed solution',
                description="How the issue could be fixed and prevented in the future",
                allow_file_upload=False,
                ),
            ),

        StringField('features',
            searchable=1,
            vocabulary="listFeatures",
            widget=MultiSelectionWidget(
                condition=showNotIn('issue_basic_edit'),
                format="checkbox",
                description="Features applying to the issue",
            ),
        ),

        ImageField('screenshot',
            widget=ImageWidget(
                 variants={ "thumb"  : (128,128),
                        "huge" : (1280,1024),},
                 description="Screenshot from the issue",
                ),
        )

    )
    #other schema items (none)
) 

# Modifications to base schema

# We don't need since its value is dynamically generated from the name and UID
schema['title'].required = False
schema['title'].widget.visible = {'edit': 'invisible', 'view': 'invisible'}


schema['description'].required = False
schema['description'].widget.visible = {'edit': 'invisible', 'view': 'invisible'}

del schema['relatedItems']

schema['id'].widget.visible = {'edit': 'invisible', 'view': 'invisible'}

# Set all read/write permissions once
for field in schema.fields():
    field.read_permission = VIEW_USABILITY_ITEMS_PERMISSION
    field.write_permission = EDIT_USABILITY_ITEMS_PERMISSION

class Issue(ATCTContent):
    """ Usability issue for an application
    
    Little magic done: 
     - Title (metadata) is generated automatically from usability id number and usability issue name
     - Issue description is stored in Dublin core description meta data field (to avoid 
       duplicate description field)
     
     I tried to use BaseContent title and description fields instead of introducing new fields in
     Issue schema.  There were problems:
     - Field order was wrong on the edit page
     - Description field didn't appear on the edit page though I modified schema 
       (edit = 'visible', isMetadata = False)
     
    If I try to delete title and description fields and then feed their values dynamically
    by overriding accessor, I get 
        # Deleting title would cause SystemError: Excessive recursion
        # in   File "E:\plone2\Data\Products\Archetypes\Field.py", line 103, in encode
        #    encoding = instance.getCharset()
    """
    
    # support disabling voting
    __implements__ = BaseContent.__implements__, IVoteable

    schema = schema

    # Allow insertion only to Application
    global_allow = False
    
    # Allow discussion by default
    allow_discussion = True

    # This form is displayed for item creation
    # default_view = "issue_add_view"
    
    meta_type      = 'Issue'
    archetype_name = 'Issue'
    content_icon   = 'usability_issue_icon.gif'
       
    # Add new action to tabs 
    actions = updateActions(ATCTContent,
            (
                {
                'id': 'view',
                'permissions': (VIEW_USABILITY_ITEMS_PERMISSION,),
                'action': 'string:${object_url}/base_view_hide_empty_fields',                             
                },                  
                {
                'id': 'edit',
                'permissions': (EDIT_USABILITY_ITEMS_PERMISSION,),
                'action': 'string:${object_url}/issue_basic_edit',
                },                             
                                
                {
                'id': 'advanced_edit',
                'name': 'Advanced Edit',
                'action': 'string:${object_url}/issue_advanced_edit',
                'permissions': (EDIT_USABILITY_ITEMS_PERMISSION,)
                },
                
                # Hide properties tab from ordinar users
                { 'id': 'metadata',
                  'name': 'Properties',
                  'action': 'string:${object_url}/base_metadata',
                  'permissions': (MANAGE_USABILITY_ITEMS_PERMISSION,),
                },   
                
                {
                  'id'          : 'local_roles',
                  'name'        : 'Sharing',                  
                  'permissions' : (MANAGE_USABILITY_ITEMS_PERMISSION,),
                },                
            )
        )
        
    aliases = updateAliases(ATCTContent,
        {
        'base_edit'      : 'issue_basic_edit',
        })        
                
    security = ClassSecurityInfo()
    
    
    security.declareProtected(ADD_ISSUES_PERMISSION, 'initializeArchetype')                        
    def initializeArchetype(self, **kwargs):
        """ Set voting enabled initially

        called by the generated addXXX factory in types tool """
        ATCTContent.initializeArchetype(self, **kwargs)
        setattr(self, "enableRatings", True)
        setattr(self, "enableVoting", True)
        
        # setup custom permissions
        
        doNotAcquirePermissions(self, (ModifyPortalContent, ManageProperties, MANAGE_USABILITY_ITEMS_PERMISSION))
        
        addPermissionsForRole(self, "Anonymous", (VIEW_USABILITY_ITEMS_PERMISSION,))
        addPermissionsForRole(self, "Member", (EDIT_USABILITY_ITEMS_PERMISSION,))
        addPermissionsForRole(self, "Manager", (ModifyPortalContent, ManageProperties, MANAGE_USABILITY_ITEMS_PERMISSION,))
        
        #addPermissionsForRole(self, "Manager", (ModifyPortalContent,))                
        # otherwise the user cannot submit edit form
        #addPermissionsForRole(self, "Member", (ModifyPortalContent,))                       
        #workflowTool = getToolByName(self, 'portal_workflow')
        #workflowTool.setStatusOf("issue_workflow", self, "inprogress")
                    
    def Title(self, **kwargs):
        return self.getId() + " " + self.getName()
                        
    def listFeatures(self):    
        """ Get list of possible taggable features from ATVocabularyManager """  
        vocabTool = getToolByName(self, 'portal_vocabularies')
        feats = vocabTool.getVocabularyByName(FEATURE_VOCABULARY).getDisplayList(self)        
        return feats
        
    security.declarePublic('isVoteable')
    def isVoteable(self):
        """ Do not allow voting of sealed item """
        workflowTool = getToolByName(self, 'portal_workflow')
        #print "Is voteable:" + str(workflowTool.getStatusOf("issue_workflow", self))
        return (workflowTool.getStatusOf("issue_workflow", self)["review_state"] == "in_progress")
        
    #
    # Modify inherited access permissions
    #
    
    security.declareProtected(VIEW_USABILITY_ITEMS_PERMISSION, 'Description')
    #def Description(self):
    #    return BaseContent.Description(self)
    
    security.declareProtected(VIEW_USABILITY_ITEMS_PERMISSION, 'Title')
    #def Title(self):
    #    return BaseContent.Title(self)
        
    #security.declareProtected(EDIT_USABILITY_ITEMS_PERMISSION, 'setTitle')
    security.declareProtected(EDIT_USABILITY_ITEMS_PERMISSION, 'setTitle')
    #def setTitle(self, value):
    #    BaseContent.setTitle(self, value)
    
    security.declareProtected(EDIT_USABILITY_ITEMS_PERMISSION, 'setDescription')
    #def setDescription(self, value):
    #    BaseContent.setDescription(self, value)
    
    # User must have rights to submit edit form (processForm checks ModifyPortalContent by default)
    security.declareProtected(EDIT_USABILITY_ITEMS_PERMISSION, 'processForm')
    

#def modify_fti(fti):
#    """ Modifies FTI for Issue object#
#
#    Makes necessary changes for Archetypes automatically generated factory type information.
#    
#    - Use custom view template to hide empty fields
#    
#    """
#    
#    for action in fti.get("actions", ()):
#        if action["id"]=="view":
#            # replace default view template
#            action["action"]="string:${object_url}/base_view_hide_empty_fields"
#            action["permissions"] = (VIEW_USABILITY_ITEMS_PERMISSION,)
#        elif action["id"]=="edit":
#            # set permissions for edit action
#            action["permissions"] = (EDIT_USABILITY_ITEMS_PERMISSION,)

registerType(Issue, PROJECTNAME)
